import json

from django.utils.module_loading import import_string

from .exceptions import LABEL_TO_EXCEPTION, SlackException
from .app_settings import app_settings


class Backend(object):
    def send(self, url, message_data):
        raise NotImplementedError()

    def validate(self, content_type, content, message_data):
        if content_type.startswith('application/json'):
            result = json.loads(content)

            if not result['ok']:
                klass = LABEL_TO_EXCEPTION.get(result['error'], SlackException)

                raise klass(result['error'], message_data)

            return result
        elif content != 'ok':
            raise SlackException(content, message_data)

        return content


def get_backend(name=None):
    """
    Wrap the backend in a function to not load it at import time.
    ``get_backend()`` caches the backend on first call.

    :param name: optional string name for backend, otherwise take from settings
    :type name: str or unicode or None
    """

    # Function's backend is initially NoneType on module import (see below)
    loaded_backend = get_backend.backend

    # Load the backend if we have a provided name, or if this function's
    # backend is still NoneType
    if name or not loaded_backend:
        loaded_backend = import_string(name or app_settings.BACKEND)()

        # If the backend hasn't been cached yet, and we didn't get a custom
        # name passed in, cache the backend
        if not (get_backend.backend or name):
            get_backend.backend = loaded_backend

    return loaded_backend


get_backend.backend = None



import json

from django.conf import settings
from django.utils.encoding import force_str
from django.template.loader import render_to_string

from .utils import get_backend
from .app_settings import app_settings


def slack_message(
    template,
    context=None,
    attachments=None,
    blocks=None,
    fail_silently=None,
    **kwargs,
):
    data = {}

    channel = kwargs.pop('channel', app_settings.CHANNEL)
    backend = get_backend(name=kwargs.pop('backend', None))

    context = dict(context or {}, settings=settings)
    if fail_silently is None:
        fail_silently = app_settings.FAIL_SILENTLY

    NOT_REQUIRED, DEFAULT_ENDPOINT, ALWAYS = range(3)

    PARAMS = {
        'text': {'default': '', 'required': NOT_REQUIRED,},  # Checked later
        'token': {
            'default': app_settings.TOKEN,
            'required': DEFAULT_ENDPOINT,
        },
        'channel': {'default': channel, 'required': DEFAULT_ENDPOINT,},
        'icon_url': {
            'default': app_settings.ICON_URL,
            'required': NOT_REQUIRED,
        },
        'icon_emoji': {
            'default': app_settings.ICON_EMOJI,
            'required': NOT_REQUIRED,
        },
        'username': {
            'default': app_settings.USERNAME,
            'required': NOT_REQUIRED,
        },
        'attachments': {
            'default': attachments,
            'render': False,
            'required': NOT_REQUIRED,
        },
        'blocks': {
            'default': blocks,
            'render': False,
            'required': NOT_REQUIRED,
        },
        'endpoint_url': {
            'default': app_settings.ENDPOINT_URL,
            'render': ALWAYS,
            'required': NOT_REQUIRED,
        },
        'as_user': {
            'default': app_settings.AS_USER,
            'render': False,
            'required': NOT_REQUIRED,
        },
    }

    for k, v in PARAMS.items():
        # First, set from default if we have one
        if v['default']:
            data[k] = v['default']

        # Render template if necessary
        if v.get('render', True):
            try:
                val = force_str(
                    render_to_string(
                        template,
                        dict(
                            context, django_slack='django_slack/{}'.format(k),
                        ),
                    ).strip()
                )
            except Exception:
                if fail_silently:
                    return
                raise

            if val:
                data[k] = val

        # Check if paramater is required
        if v['required'] == ALWAYS:
            if data.get(k, None):
                continue

            if fail_silently:
                return

            raise ValueError(
                "Missing or empty required parameter: {}".format(k)
            )

    if 'text' not in data and 'attachments' not in data:
        raise ValueError(
            "text parameter is required if attachments is not set",
        )

    # Ensure that as_user is either "true" or not present (rather than "True"
    # or "False", etc.).
    #
    # This also prevents an encoding error under (just) Django 2.1 due to an
    # upstream regression (<https://github.com/lamby/django-slack/issues/85>).
    #
    if data.pop('as_user', app_settings.AS_USER):
        data['as_user'] = 'true'

    # The endpoint URL is not part of the data payload but as we construct it
    # within `data` we must remove it.
    endpoint_url = data.pop('endpoint_url')

    # If a custom endpoint URL was specified then we need to wrap it, otherwise
    # we need to ensure attachments are encoded.
    if endpoint_url == app_settings.DEFAULT_ENDPOINT_URL:
        # Check parameters that are only required if we don't specify a custom
        # endpoint URL.
        for k, v in PARAMS.items():
            if v['required'] != DEFAULT_ENDPOINT:
                continue

            if not data.get(k, None):
                if fail_silently:
                    return

                raise ValueError(
                    "{} parameter is required if custom endpoint URL is not "
                    "specified".format(k),
                )

        for x in ('attachments', 'blocks'):
            if x in data:
                data[x] = json.dumps(data[x])
    else:
        data = {'payload': json.dumps(data)}

    try:
        return backend.send(endpoint_url, data, **kwargs)
    except Exception:
        if not fail_silently:
            raise



import json

from django.test import TestCase

from django_slack.exceptions import ChannelNotFound, MsgTooLong
from django_slack.backends import Backend


class TestOverride(TestCase):
    def test_ok_result(self):
        backend = Backend()
        backend.validate('application/json', json.dumps({'ok': True}), {})

    def test_msg_too_long_result(self):
        # Arbitrarily chosen 'simple' error
        backend = Backend()
        with self.assertRaisesRegex(
            MsgTooLong, r"msg_too_long",
        ):
            backend.validate(
                'application/json',
                json.dumps({'ok': False, 'error': 'msg_too_long'}),
                {},
            )

    def test_channel_not_found_result(self):
        backend = Backend()
        with self.assertRaisesRegex(
            ChannelNotFound, r"channel 'bad-channel' could not be found",
        ):
            backend.validate(
                'application/json',
                json.dumps({'ok': False, 'error': 'channel_not_found'}),
                {'channel': 'bad-channel'},
            )



class SlackException(ValueError):
    def __init__(self, message, message_data):
        super(SlackException, self).__init__(message)
        self.message_data = message_data


class ChannelNotFound(SlackException):
    def __str__(self):
        # Override base __str__ to ensure we include the channel name in the
        # error message
        return u"{}: channel '{}' could not be found".format(
            self.__class__.__name__, self.message_data['channel'],
        )


class IsArchived(SlackException):
    pass


class FatalError(SlackException):
    pass


class MsgTooLong(SlackException):
    pass


class NoText(SlackException):
    pass


class RateLimited(SlackException):
    pass


class NotAuthed(SlackException):
    pass


class InvalidAuth(SlackException):
    pass


class TokenRevoked(SlackException):
    pass


class AccountInactive(SlackException):
    pass


class UserIsBot(SlackException):
    pass


LABEL_TO_EXCEPTION = {
    'channel_not_found': ChannelNotFound,
    'is_archived': IsArchived,
    'fatal_error': FatalError,
    'msg_too_long': MsgTooLong,
    'no_text': NoText,
    'rate_limited': RateLimited,
    'not_authed': NotAuthed,
    'invalid_auth': InvalidAuth,
    'token_revoked': TokenRevoked,
    'account_inactive': AccountInactive,
    'user_is_bot': UserIsBot,
}



#!/usr/bin/env python3

from setuptools import setup, find_packages

setup(
    name='django-slack',
    url="https://chris-lamb.co.uk/projects/django-slack",
    version='5.16.0',
    description="Provides easy-to-use integration between Django projects and "
    "the Slack group chat and IM tool.",
    author="Chris Lamb",
    author_email="chris@chris-lamb.co.uk",
    license="BSD-3-Clause",
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.5",
    install_requires=('Django>=2', 'requests'),
)
